--[[-------------------------------------------------------------------------
	InterruptHelper
	Copyright (C) 2011  Morsker
	All Rights Reserved.
	Please contact Morsker through PM on wowace.com.

	Helps with interrupt notifications and announcements.
---------------------------------------------------------------------------]]
local addonName = ...
local InterruptHelper = LibStub("AceAddon-3.0"):NewAddon(addonName, "AceEvent-3.0")
_G.InterruptHelper = InterruptHelper

--------------
--  Locals  --
--------------

-- GLOBALS: _G
-- GLOBALS: InterruptHelperDB
-- GLOBALS: EnableAddOn
-- GLOBALS: LoadAddOn
-- GLOBALS: LibStub
-- GLOBALS: CreateFrame
-- GLOBALS: InterfaceOptionsFrameAddOnsButton1
-- GLOBALS: InterfaceOptionsFrame_OpenToCategory
-- GLOBALS: InterfaceOptions_AddCategory
local GetSpellLink = GetSpellLink
local GetSpellInfo = GetSpellInfo
local GetZonePVPInfo = GetZonePVPInfo
local UnitCanAttack = UnitCanAttack
local UnitCastingInfo = UnitCastingInfo
local UnitChannelInfo = UnitChannelInfo
local UnitGUID = UnitGUID
local UnitInBattleground = UnitInBattleground
local UnitInRaid = UnitInRaid
local UnitIsDead = UnitIsDead
local UnitName = UnitName
local GetNumPartyMembers = GetNumPartyMembers
local GetChannelName = GetChannelName
local PlaySoundFile = PlaySoundFile
local SendChatMessage = SendChatMessage
local select = select
local format = format
local gsub = gsub
local pairs = pairs
local print = print
local strlower = strlower
local type = type

local L = LibStub:GetLibrary("AceLocale-3.0"):GetLocale("InterruptHelper")
local LSM = LibStub("LibSharedMedia-3.0")
local DBVERSION = 1
local db
local cleuFrame = CreateFrame("Frame", "InterruptHelperCLEU"); cleuFrame:Hide();
local castFrame = CreateFrame("Frame", "InterruptHelperCast"); castFrame:Hide();
local flashFrame
local flashFrameTexture
local playerGUID
local targetGUID
local focusGUID
local lastSayFormat
local sayFormatInternal
local targetIsWarning
local focusIsWarning

InterruptHelper.definedSounds = {
	["Simon Chime"] = [[Sound\Doodad\SimonGame_LargeBlueTree.ogg]],
	["Simon Chime 2"] = [[Sound\Doodad\SimonGame_SmallBlueTree.ogg]],
	["Short Circuit"] = [[Sound\Spells\SimonGame_Visual_BadPress.ogg]],
	["Frost Explosion"] = [[Sound\Spells\Sindragosa_Xplosion_Frost_Impact_03.ogg]],
	["Squeek"] = [[Sound\Creature\ArcaneTitan\ArcaneTitan_Wound_01.ogg]],
	["Exploding Stone"] = [[Sound\Doodad\GO_ExplodingStoneImpact01_01.ogg]],
	["Zap Thud"] = [[Sound\Doodad\GO_TBDestroyTowerFX_DomeImpact_03.ogg]],
	["Bubble Blast"] = [[Sound\Spells\FX_BubbleBlast_Impact_06.ogg]],
	["Heroic Strike"] = [[Sound\Spells\SPELL_HeroicStrike_Impact_02.ogg]],
	["Shadow Strike"] = [[Sound\Spells\ShadowStrike_Impact_Chest.ogg]],
	["Tech Gun 1"] = [[Sound\Spells\SPELL_ShootTechGun_Cast_05.ogg]],
	["Tech Gun 2"] = [[Sound\Spells\SPELL_ShootTechGun_Impact_04.ogg]],
	["Throw Gears"] = [[Sound\Spells\SPELL_ThrowGears_Impact_03.ogg]],
	["Frost Impact"] = [[Sound\Spells\SPELL_MA_CurtainFrost_Impact_03.ogg]],
	["Colossus Smash"] = [[Sound\Spells\SPELL_WR_ColossusSmash_Impact_02.ogg]],
	["Snowball"] = [[Sound\Spells\SnowballPowdery_Impact_Base.ogg]],
	["Hearthstone Bind"] = [[Sound\Spells\bind2_Impact_Base.ogg]],
	["Wallow"] = [[Sound\Spells\Spell_Wallow_Impact_05.ogg]],
}
for k,v in pairs(InterruptHelper.definedSounds) do
	LSM:Register("sound", k, v)
end
InterruptHelper.definedSounds["None"] = LSM:Fetch("sound", "None")

local defaults = {
	profile = {
		annAnnounce = true,
		annSayFormat = "Interrupted: $S",
		annSolo = "TEXT",
		annParty = "PARTY",
		annRaid = "RAID",
		annBattleground = "TEXT",
		annCombatZone = "TEXT",

		warnSimultaneous = "focus",

		targetEnabled = true,
		targetSound = "Zap Thud",
		targetFlashTime = 0.25,
		targetFlashR = 1.00,
		targetFlashG = 1.00,
		targetFlashB = 0.25,
		targetFlashA = 0.3,

		focusEnabled = true,
		focusSound = "Tech Gun 1",
		focusFlashTime = 0.25,
		focusFlashR = 0.655,
		focusFlashG = 0.341,
		focusFlashB = 0.839,
		focusFlashA = 0.3,

		tracking = {
		},
		trackLearn = true,
		trackWarnUnknown = false,
		trackWarnAll = false,
	}
}

function InterruptHelper:PushUpvalues()
end

function InterruptHelper:ZoneCheck()
	local pvpType = GetZonePVPInfo()
	self.combatZone = not not (pvpType == "combat")
end

function InterruptHelper:PLAYER_ENTERING_WORLD()
	targetIsWarning = false
	focusIsWarning = false
	playerGUID = UnitGUID("player")
	targetGUID = UnitGUID("target")
	focusGUID = UnitGUID("focus")
	self:ZoneCheck()
end

function InterruptHelper:PLAYER_TARGET_CHANGED(event)
	targetGUID = UnitGUID("target")
	return self.CastEventHandler(castFrame, event, "target")
end

function InterruptHelper:PLAYER_FOCUS_CHANGED(event)
	focusGUID = UnitGUID("focus")
	return self.CastEventHandler(castFrame, event, "focus")
end

function InterruptHelper:FlashScreen(time, r, g, b, a)
	if time <= 0 then return end
	if not flashFrame then
		flashFrame = CreateFrame("Frame", "InterruptHelperFlash")
		flashFrame.timeLeft = 0
		flashFrame:SetFrameStrata("BACKGROUND")
		flashFrameTexture = flashFrame:CreateTexture(nil, "BACKGROUND")
		flashFrameTexture:SetTexture("Interface\\Addons\\InterruptHelper\\Textures\\Solid")
		flashFrameTexture:SetAllPoints(true)
		flashFrame:SetAllPoints(true)
		flashFrame:Hide()

		local function OnUpdate(self, elapsed)
			local timeLeft = self.timeLeft - elapsed
			self.timeLeft = timeLeft
			if timeLeft <= 0 then
				self:Hide()
			end
		end
		flashFrame:SetScript("OnUpdate", OnUpdate)
	end
	flashFrame.timeLeft = time
	flashFrameTexture:SetVertexColor(r, g, b, a)
	flashFrame:Show()
end

function InterruptHelper:Warn(unitID)
	PlaySoundFile(LSM:Fetch("sound", db.profile[unitID.."Sound"]), "Master")
	self:FlashScreen(
		db.profile[unitID.."FlashTime"],
		db.profile[unitID.."FlashR"],
		db.profile[unitID.."FlashG"],
		db.profile[unitID.."FlashB"],
		db.profile[unitID.."FlashA"])
end

function InterruptHelper:SayText(msg)
	local context
	if UnitInBattleground("player") then
		context = "annBattleground"
	elseif UnitInRaid("player") then
		if self.combatZone then
			context = "annCombatZone"
		else
			context = "annRaid"
		end
	elseif GetNumPartyMembers() > 0 then
		context = "annParty"
	else
		context = "annSolo"
	end
	local chatType = db.profile[context]
	if chatType == "TEXT" then
		print(msg)
	elseif chatType ~= "NONE" then
		local channel
		if chatType == "WHISPER" then
			channel = db.profile[context.."Player"]
		elseif chatType == "CHANNEL" then
			local channelName = strlower(db.profile[context.."Channel"])
			local i,_,iChannelName = 1, GetChannelName(1)
			while iChannelName do
				if strlower(iChannelName) == channelName then
					channel = i
					break
				end
				i,_,iChannelName = i + 1, GetChannelName(i + 1)
			end
			if channel == nil then
				return
			end
		end
		SendChatMessage(msg, chatType, nil, channel)
	end
end

function InterruptHelper.CLEUHandler(cleuFrame, _, timestamp, event, ...)
	if event ~= "SPELL_INTERRUPT" then return end
	local sourceGUID = ...
	if sourceGUID ~= playerGUID then return end

	local spellID = select(10, ...)
	if lastSayFormat ~= db.profile.annSayFormat then
		sayFormatInternal = db.profile.annSayFormat
		sayFormatInternal = gsub(sayFormatInternal, "%%", "%%%%")
		sayFormatInternal = gsub(sayFormatInternal, "$s", "%%1$s")
		sayFormatInternal = gsub(sayFormatInternal, "$S", "%%2$s")
		lastSayFormat = db.profile.annSayFormat
	end
	local spellName = GetSpellInfo(spellID)
	local spellLink = GetSpellLink(spellID)
	local msg = format(sayFormatInternal, spellName, spellLink)
	return InterruptHelper:SayText(msg)
end

function InterruptHelper.CastEventHandler(castFrame, event, unitID)
	local isWarning
	if unitID == "target" then
		isWarning = targetIsWarning
	elseif unitID == "focus" then
		isWarning = focusIsWarning
	else
		return
	end

	local shouldWarn = false
	local castingSpell
	if (event == "UNIT_SPELLCAST_STOP" or event == "UNIT_SPELLCAST_CHANNEL_STOP" or event == "UNIT_SPELLCAST_NOT_INTERRUPTIBLE") then
		-- Do nothing. Leave shouldWarn as false by default.
	else
		local hostileTarget = UnitCanAttack("player", unitID) and not UnitIsDead(unitID)
		if hostileTarget then
			local _,immuneToInterrupts
			castingSpell,_,_,_,_,_,_,_,immuneToInterrupts = UnitCastingInfo(unitID)
			if not castingSpell then
				castingSpell,_,_,_,_,_,_,immuneToInterrupts = UnitChannelInfo(unitID)
			end

			if castingSpell and not immuneToInterrupts then
				shouldWarn = true
			end
		end -- else, do nothing, leaving shouldWarn false by default.
	end

	if shouldWarn ~= isWarning then
		-- Target or focus?
		if unitID == "target" then
			targetIsWarning = shouldWarn
		else
			focusIsWarning = shouldWarn
		end
		-- On or off?
		if shouldWarn then
			-- Don't double-warn if it's both target & focus.
			if (targetGUID ~= focusGUID) or (unitID == db.profile.warnSimultaneous) then
				-- Check settings
				local unitName = UnitName(unitID)
				local tracking = db.profile.tracking[unitName]
				if tracking then
					local setting = tracking[castingSpell]
					if setting == nil then
						setting = (tracking._warnUnknown ~= nil) and tracking._warnUnknown or db.profile.trackWarnUnknown
						local learn = (tracking._learn ~= nil) and tracking._learn or db.profile.trackLearn
						if learn then
							tracking[castingSpell] = setting
							if InterruptHelper.options then
								InterruptHelper:NewTrackingSpellView(unitName, castingSpell)
								InterruptHelper:OrderTrackingView()
							end
						end
					end
					if setting then
						if db.profile[unitID.."Enabled"] then
							InterruptHelper:Warn(unitID)
						end
					end
				elseif db.profile.trackWarnAll then
					-- Minor code duplication. Haven't found a good way to consolodate yet.
					if db.profile[unitID.."Enabled"] then
						InterruptHelper:Warn(unitID)
					end
				end
			end
		else
			-- Stop?
		end
	end
end

local function LoadOptions()
	if InterruptHelper.options then return end
	EnableAddOn("InterruptHelper_Options")
	LoadAddOn("InterruptHelper_Options")
	InterruptHelper:InitOptions()
end

-- GLOBALS: SLASH_INTERRUPTHELPER1
-- GLOBALS: SLASH_INTERRUPTHELPER2
-- GLOBALS: SLASH_INTERRUPTHELPER3
SLASH_INTERRUPTHELPER1 = "/ih"
SLASH_INTERRUPTHELPER2 = "/ihelper"
SLASH_INTERRUPTHELPER3 = "/interrupthelper"
function SlashCmdList.INTERRUPTHELPER()
	LoadOptions()
	local ACD = LibStub("AceConfigDialog-3.0")
	local targetName = UnitName("target")
	ACD:Open("InterruptHelper")
	if targetName and db.profile.tracking[targetName] then
		ACD:SelectGroup("InterruptHelper", "tracking", targetName)
	end
end

function InterruptHelper:OnInitialize()
	self.OnInitialize = nil
	InterruptHelperDB = InterruptHelperDB or { dbversion = DBVERSION }
	db = LibStub("AceDB-3.0"):New("InterruptHelperDB", defaults, true)
	InterruptHelper.db = db
	self:PushUpvalues()

	-- UI stub for opening options, but only if we need it.
	if not InterruptHelper.options then
		local dummyOptionsFrame = CreateFrame("Frame")
		dummyOptionsFrame.name = "InterruptHelper"
		dummyOptionsFrame:Hide()
		dummyOptionsFrame:SetScript("OnShow", function(frame)
			frame:SetScript("OnShow", nil)
			frame:Hide()
			LoadOptions()
			InterfaceOptionsFrame_OpenToCategory("InterruptHelper")
			local i, button = 1, InterfaceOptionsFrameAddOnsButton1
			while button and button:IsVisible() do
				if button.element.name == "InterruptHelper" and button.element.hasChildren then
					button.toggle:Click()
					return
				end
				i, button = i + 1, _G["InterfaceOptionsFrameAddOnsButton"..(i + 1)]
			end
		end)
		InterfaceOptions_AddCategory(dummyOptionsFrame, "InterruptHelper")
	end

	playerGUID = UnitGUID("player")

	self.cleuFrame = cleuFrame
	cleuFrame:SetScript("OnEvent", self.CLEUHandler)
	self.castFrame = castFrame
	castFrame:SetScript("OnEvent", self.CastEventHandler)
end

function InterruptHelper:OnEnable()
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_FOCUS_CHANGED")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "ZoneCheck")
	self:ZoneCheck()

	if db.profile.annAnnounce then
		cleuFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	end

	castFrame:RegisterEvent("UNIT_SPELLCAST_START")
	castFrame:RegisterEvent("UNIT_SPELLCAST_STOP")
	castFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START")
	castFrame:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP")
	castFrame:RegisterEvent("UNIT_SPELLCAST_INTERRUPTIBLE")
	castFrame:RegisterEvent("UNIT_SPELLCAST_NOT_INTERRUPTIBLE")
end

function InterruptHelper:OnDisable()
	cleuFrame:UnregisterAllEvents()
	castFrame:UnregisterAllEvents()
end
